package ru.bitel.bgbilling.modules.tv.dyn.lfstrm;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.sql.Connection;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import bitel.billing.server.contract.bean.Contract;
import bitel.billing.server.contract.bean.ContractManager;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.json.JSONObject;

import ru.bitel.bgbilling.common.BGException;
import ru.bitel.bgbilling.kernel.container.managed.ServerContext;
import ru.bitel.bgbilling.kernel.module.common.bean.User;
import ru.bitel.bgbilling.modules.tv.common.bean.TvAccount;
import ru.bitel.bgbilling.modules.tv.server.bean.TvAccountDao;
import ru.bitel.bgbilling.modules.tv.server.handler.HttpHandler;
import ru.bitel.bgbilling.server.util.Setup;
import ru.bitel.common.TimeUtils;
import ru.bitel.common.Utils;
import ru.bitel.common.model.Tied;
import ru.bitel.oss.systems.inventory.product.common.bean.Product;
import ru.bitel.oss.systems.inventory.product.common.bean.ProductSpec;
import ru.bitel.oss.systems.inventory.product.common.bean.ProductSpecActivationMode;
import ru.bitel.oss.systems.inventory.product.common.service.ProductService;
import ru.bitel.oss.systems.inventory.product.server.bean.ProductDao;
import ru.bitel.oss.systems.inventory.product.server.bean.ProductSpecDao;
import ru.bitel.oss.systems.order.product.common.service.ProductOrderService;
import ru.bitel.oss.systems.order.product.server.service.ProductOrderServiceImpl;

public class LifeStreamHandler
    implements HttpHandler
{
    private static final Logger logger = LogManager.getLogger();

    @Override
    public void handle( int moduleId, HttpServletRequest request, HttpServletResponse response )
    {
        ServerContext context = new ServerContext( Setup.getSetup(), moduleId, User.USER_SERVER );
        ServerContext.push( context );

        Setup setup = Setup.getSetup();

        try(Connection connection = setup.getDBConnectionFromPool())
        {
            request.setCharacterEncoding( "UTF-8" );
            response.setCharacterEncoding( "UTF-8" );

            JSONObject upsaleRequestV2 = parseUpsaleRequestV2( request );

            // ищем клиента
            TvAccount tvAccount;
            try( TvAccountDao tvAccountDao = new TvAccountDao( connection, context.getModuleId() ) )
            {
                tvAccount = tvAccountDao.getByDeviceAccountId( upsaleRequestV2.optString( "id" ), new Date() );
                if ( tvAccount == null )
                {
                    sendErrorResponse400( response, "error_no_account", "id = " + upsaleRequestV2.optString( "id" ), "" );
                    return;
                }
            }

            // ищем продукт
            ProductSpec productSpec;
            try ( ProductDao productDao = new ProductDao( connection, User.USER_SERVER );
                  ProductSpecDao productSpecDao = new ProductSpecDao( connection ) )
            {
                // продукт, который хотят подключить
                productSpec = productSpecDao.getByIdentifier( upsaleRequestV2.optString( "offerId" ) );
                if ( productSpec == null )
                {
                    sendErrorResponse400( response, "error_bad_request", "Продукт с Identifier = " + upsaleRequestV2.optString( "offerId" ) + "[#offerId] не найдена", "" );
                    return;
                }

                // проверка не подключен ли он уже у клиента
                // получаем список текущих услуг на аккаунте
                List<Product> productList = productDao.list( moduleId, tvAccount.getContractId(), tvAccount.getId(), false, new Date() );
                for ( Product product : productList )
                {
                    if ( product.getProductSpecId() == productSpec.getId() )
                    {
                        sendErrorResponse400( response, "no_action_required", "", "" );
                        return;
                    }
                }

                //проверка, что продукт сов

                //проверяем возможно ли активировать продукт на договоре
                List<ProductSpecActivationMode> activationModes = productSpec.getActivationModeList();
                if( Utils.isEmptyCollection( activationModes ) )
                {
                    String errorText = "Для продукта с Identifier = " + upsaleRequestV2.optString( "offerId" ) + " не задано ни одного режима активации!";
                    logger.error( errorText );
                    sendErrorResponse400( response, "error_bad_request", errorText, "" );
                    return;
                }

                if( activationModes.size() > 1 )
                {
                    String errorText = "Для продукта с Identifier = " + upsaleRequestV2.optString( "offerId" ) + " задано более одного режима активации!";
                    logger.error( errorText );
                    sendErrorResponse400( response, "error_bad_request", errorText, "" );
                    return;
                }

                //время активации продукта - устанавливаем текущее время
                Date activationTime = new Date( TimeUtils.rountToSeconds( System.currentTimeMillis() ) );

                ProductSpecActivationMode activationMode = activationModes.get( 0 );
                if( !TimeUtils.dateInRange( activationTime, activationMode.getDateFrom(), activationMode.getDateTo() ) )
                {
                    String errorText = "Активация продукта с ProductId=" + productSpec.getId() + " запрещена с заданным режимом активации запрещена!";
                    logger.error( errorText );
                    sendErrorResponse400( response, "error_bad_request", errorText, "" );
                    return;
                }

                //создание продукта и его активация
                int contractId = tvAccount.getContractId();
                Contract contract;
                try( ContractManager contractManager = new ContractManager( connection ) )
                {
                    contract = contractManager.getContractById( contractId );
                    if( contract == null )
                    {
                        String errorText = "Для продукта с Identifier = " + upsaleRequestV2.optString( "offerId" ) + " не найден договор с ID " + contractId;
                        logger.error( errorText );
                        sendErrorResponse400( response, "error_bad_request", errorText, "" );
                        return;
                    }
                }

                if( !checkIncompatible( moduleId, contractId, tvAccount.getId(), productSpec, context ) )
                {
                    String errorText = "Продукта с Identifier = " + upsaleRequestV2.optString( "offerId" ) + " несовместим с продуктами на договоре";
                    logger.error( errorText );
                    sendErrorResponse400( response, "error_bad_request", errorText, "" );
                    return;
                }

                Product product = new Product();
                product.setAccountId( tvAccount.getId() );
                product.setContractId( contractId );
                product.setProductSpecId( productSpec.getId() );
                product.setActivationTime( activationTime );
                product.setActivationModeId( activationMode.getId() );
                product.setDeviceState( Product.STATE_ENABLED );
                product.setUserId( User.USER_SERVER );
                product.setTimeFrom( activationTime );

                //настраиваем период продукта
                ProductOrderService productOrderService = context.getService( ProductOrderService.class, moduleId );
                ProductService productService = context.getService( ProductService.class, moduleId );

                productOrderService.adjustProductPeriod( product, activationMode, false );

                List<Product> activeProductList = productOrderService.activeProductList( contractId, tvAccount.getId(), activationTime );
                final Date _productTimeTo = product.getTimeTo();
                try
                {
                    product.setTimeTo( new Date( product.getTimeFrom().getTime() + 1000L ) );
                    ((ProductOrderServiceImpl)productOrderService).getTieUtils().checkAdd( activeProductList, product, Calendar.MILLISECOND );
                }
                catch( Tied.TieUnresolvedException ex )
                {
                    String errorText = "Невозможно активировать продукт с Identifier = " + upsaleRequestV2.optString( "offerId" ) + " т.к. на договоре есть уже активные продукты";
                    logger.error( errorText );
                    sendErrorResponse400( response, "error_bad_request", errorText, "" );
                    return;
                }
                finally
                {
                    product.setTimeTo( _productTimeTo );
                }

                // Если всё нормально, то активируем продукт на договоре
                logger.info( "Активация продукта для договора contractId=" + contractId );
                int productId = productService.productUpdate( product );
                productOrderService.productActivate( product, new Date(), false, false );
                context.commit();
                logger.info( "Активирован продукт ProductId=" + productId + ", для договора contractId=" + contractId );

                //сообщаем, что успешно активировали продукт
                sendRequestProcessedResponse( response, "subscription_added", null, 0, 0 );
            }
        }
        catch( Exception ex )
        {
            logger.error( ex.getMessage(), ex );
            
            sendErrorResponse500( response, "exception", ex.getLocalizedMessage(), ex.toString() );
        }
    }

    private boolean checkIncompatible( int moduleId, int contractId, int accountId, ProductSpec newProduct, ServerContext context )
        throws BGException
    {
        ProductService productService = context.getService( ProductService.class, 0 );

        if( newProduct == null )
            return true;

        //все продукты на данный момент времени для этого аккаунта
        List<Product> productsOnContract = productService.productList( moduleId, contractId, accountId, false, null, null, new Date(), new Date(), true, true );

        Set<Integer> incompatible = newProduct.getIncompatible();
        if( Utils.isEmptyCollection( incompatible ) )
            return true;

        return productsOnContract.stream()
                                 .mapToInt( Product::getProductSpecId )
                                 .boxed()
                                 .collect( Collectors.toList())
                                 .stream()
                                 .noneMatch( incompatible::contains );
    }
    
    private JSONObject parseUpsaleRequestV2( HttpServletRequest request )
        throws Exception
    {
        return new JSONObject( new String( Utils.readByBlock( request.getInputStream() ), "UTF-8" ) );
    }
    
    /**
     * Результат выполнения запроса. Может принимать одно из следующих значений:
     *   subscription_added - подключена дополнительная подписка, детали в поле transaction
     *   subscription_changed - изменена базовая подписка, детали в поле transaction
     *   no_action_required - подписка уже существует и активна
     *   no_sufficient_balance - недостаточно средств для подключения, сумма для пополнения в поле refillAmount
     *   no_subscription_rules - подключение подписки противоречит внутренним правилам биллинга
     * @param response
     * @param id
     */
    private void sendRequestProcessedResponse( HttpServletResponse response, String id, String transactionId, long transactionTimestamt, int refillAmount )
    {
        JSONObject responseJson = new JSONObject();
        responseJson.put( "id", id );
        if ( transactionId != null )
        {
            JSONObject transaction = new JSONObject();
            transaction.put( "id", transactionId );
            transaction.put( "timestamp", transactionTimestamt );
            responseJson.put( "transaction", transaction );
        }
        if ( refillAmount > 0 )
        {
            responseJson.put( "refillAmount", refillAmount );
        }
        
        response.setContentType( "application/json" );
        try( Writer writer = new OutputStreamWriter( response.getOutputStream(), "UTF-8" ) )
        {
            writer.write( responseJson.toString() );
            writer.flush();
        }
        catch( IOException ex )
        {
            logger.error( ex );
        }
    }
    
    /**
     * В поле id объекта errorResponse должен быть один из следующих идентификаторов:
     *   error_no_account - абонента нет в биллинге оператора
     *   error_failed_activation - не удалось активировать услугу на TV-платформе, в поле message должно быть значение из поля error ответа от TV платформы
     *   error_bad_request - ошибка в формате запроса или содержании его полей
     * @param id
     * @return
     */
    private void sendErrorResponse400( HttpServletResponse response, String id, String message, String details )
    {
        try
        {
            response.setContentType( "application/json" );
            response.sendError( HttpServletResponse.SC_BAD_REQUEST, getErrorResponse( id, message, details ) );
        }
        catch( IOException ex )
        {
            logger.error( ex );
        }
    }

    private void sendErrorResponse500( HttpServletResponse response, String id, String message, String details )
    {
        try
        {
            response.setContentType( "application/json" );
            response.sendError( HttpServletResponse.SC_INTERNAL_SERVER_ERROR, getErrorResponse( id, message, details ) );
        }
        catch( IOException ex )
        {
            logger.error( ex );
        }
    }
    
    private String getErrorResponse( String id, String message, String details )
    {
        JSONObject errorJson = new JSONObject();
        errorJson.put( "id", id );
        errorJson.put( "message", message );
        errorJson.put( "details", details );
        return errorJson.toString();
    }
}
